'''Contains helpers to access AESO's Available Transfer Capacity report
at <http://itc.aeso.ca/itc/public/atcQuery.do>.  This report lists
transfer capacities over intertie links at given points in time.

AESO notes that these ATC limits might be further imapcted by conditions
on the British Columbia or Saskatchewan grids.  Interested parties
should visit B.C. Hydro's OASIS website
<http://www.oatioasis.com/cwo_default.htm> and SaskPower's OASIS website
<http://www.oatioasis.com/spc/index.html>.'''

########################################################################
## Standard library imports
import csv
from time import strptime
from time import mktime
from datetime import datetime
from datetime import date
from datetime import timedelta
from decimal import Decimal
import urllib
import urllib2
import shutil
import time


########################################################################
## Import 3rd party modules
import pytz


########################################################################
## AESO Modules
from aeso import AB_TZ
from aeso._util import DayBlockIt


def urlopen_atc(start_date, end_date):
    '''Returns a file containing the available transfer capcity report.
    Using this query, Alberta ATC data can be obtained for historical
    dates starting Nov 22, 1999, and for forecasted dates up to six
    months into the future (AESO,
    http://itc.aeso.ca/itc/public/atcQuery.do, retrieved 2010-11-14).
    Due to limitations of the AESO source query end_date - start_date
    must not exceed six months.

    :param start_date: :class:`datetime.date`
    :param end_date: :class:`datetime.date`
    :rtype: file-like object as returned by urlopen.

    .. versionadded:: 0.6

    Usage example::

        >>> # Standard python libraries
        >>> from datetime import date
        >>> from datetime import timedelta
        >>>
        >>> # 3rd Party Libraries
        >>> from aeso import atc
        >>>
        >>> end_date = date.today()
        >>> start_date = end_date - timedelta(1)
        >>>
        >>> f = atc.urlopen_atc(start_date, end_date)
        >>> text = f.read()
        >>> f.close()

    .. note::

        AESO's raw ATC report can be accessed at
        <http://itc.aeso.ca/itc/public/atcQuery.do>.'''

    DATE_FORMAT = '%Y-%m-%d'

    s_from_epoch = time.time() + 5*30*24*60*60 # ~5 months
    ms_from_epoch = 1000. * s_from_epoch
    utc_dt = time.gmtime(s_from_epoch)
    local_dt = time.localtime(s_from_epoch)

    LOCAL_DT_FORMAT = '%Y-%m-%d %H:%M:%s.000 %Z'
    UTC_DT_FORMAT = '%Y-%m-%d %H:%M:%s.000 UTC'

    url = 'http://itc.aeso.ca/itc/public/atcQuery.do'
    parameters = {
        'fileFormat' : 'publicCsvFormatter',
        'startDate' : start_date.strftime(DATE_FORMAT),
        'endDate' : end_date.strftime(DATE_FORMAT),
        # Commenting these next two lines causes internal web server errors! (2009-12-14)
        'availableEffectiveDate' : '943279200000 1999-11-22 07:00:00.000 MST (1999-11-22 14:00:00.000 GMT)',
        'availableExpiryDate' : '%d %s (%s)' % (ms_from_epoch, time.strftime(LOCAL_DT_FORMAT, local_dt), time.strftime(UTC_DT_FORMAT, utc_dt)),
        '' : 'Submit',
    }

    encoded_params = urllib.urlencode(parameters)
    #http://itc.aeso.ca/itc/public/atcQuery.do?fileFormat=publicCsvFormatter&startDate=2010-11-10&endDate=2010-11-25&availableEffectiveDate=943279200000+1999-11-22+07%3A00%3A00.000+MST+%281999-11-22+14%3A00%3A00.000+GMT%29&availableExpiryDate=1305439200000+2011-05-15+00%3A00%3A00.000+MDT+%282011-05-15+06%3A00%3A00.000+GMT%29&=Submit
    #print 'opening url: %s?%s' % (url, encoded_params)
    f = urllib2.urlopen(url, encoded_params)

    return f


def dump_atc(f, start_date, end_date):
    '''Download available transfer capacity data between *start_date* and
    *end_date* and write it to file-object *f*.  Unlike urlopen_atc,
    there is no restriction on the amount of data available.  An
    iterator is used to query data in six-month blocks before it is
    written to *f*.

    :param f: file-like object.
    :param start_date: :class:`datetime.date`
    :param end_date: :class:`datetime.date`

    .. versionadded:: 0.6

    Usage example::

        >>> # Standard python libraries
        >>> from tempfile import TemporaryFile
        >>>
        >>> # 3rd Party Libraries
        >>> from aeso import atc
        >>>
        >>> end_date = date.today()
        >>> start_date = end_date - timedelta(1)
        >>>
        >>> f = TemporaryFile()
        >>> atc.dump_atc(f, start_date, end_date)
        >>> f.close()'''


    for (start, end) in DayBlockIt(start_date, end_date, 30*6):
        # print 'From', start, 'to', end
        src = urlopen_atc(start, end)
        shutil.copyfileobj(src, f)

        if end < end_date:
            time.sleep(2)


def parse_atc_file(f):
    '''Yields :class:`TransferCapacity` objects as extracted from the open
    file-object *f*.  File *f* should contain an available transfer
    capacity report in CSV format.

    :param f: file-like object.

    .. versionadded:: 0.6

    Usage example::

        >>> # Standard library imports
        >>> from StringIO import StringIO
        >>> import datetime
        >>>
        >>> # 3rd Party Libraries
        >>> from aeso import atc
        >>>
        >>> end_date = date.today()
        >>> start_date = end_date - timedelta(1)
        >>>
        >>> f = StringIO()
        >>> atc.dump_atc(f, start_date, end_date)
        >>> f.seek(0)
        >>> data = list(atc.parse_atc_file(f))
        >>> f.close()
        >>>
        >>> # *data* now contains a list of TransferCapacity objects.'''

    #~ Posted As Of: "2009-12-14 16:01:42
    #~ "Date","Hour Ending","BC Export ATC","BC Export Reason","BC Import ATC","BC Import Reason","SK Export ATC","SK Export Reason","SK Import ATC","SK Import Reason",
    #~ 2009-01-01,1,330,"OPP 521 Limit",525,"Opp 312 Limit",35,"",153,""
    #~ 2009-01-01,2,360,"OPP 521 Limit",500,"Opp 312 Limit",35,"",153,""

    date_posted = f.readline()
    column_headings = f.readline()

    # Header lines look like this:
    # "Posted As Of: "2009-12-15 19:36:43
    # "Date","Hour Ending","BC Export ATC","BC Export Reason","BC Import ATC","BC Import Reason","SK Export ATC","SK Export Reason","SK Import ATC","SK Import Reason",

    reader = csv.reader(f)
    for idx, line in enumerate(reader):
        # Ignore header lines.   This is done so that files can be concatenated.
        if line[0].startswith('Posted'):
            # Ignore first header line
            continue
        elif line[0] == 'Date':
            # Ignore second header line
            continue

        utc_dt = _normalize_atc_dtstr_to_utc(line)

        bc_export_power = line[2]
        bc_export_reason = line[3]
        bc_import_power = line[4]
        bc_import_reason = line[5]
        bc_atc_limit = AtcLimit(AtcLink.BC, bc_import_power, bc_import_reason, bc_export_power, bc_export_reason)

        sk_export_power =  line[6]
        sk_export_reason = line[7]
        sk_import_power = line[8]
        sk_import_reason = line[9]
        sk_atc_limit = AtcLimit(AtcLink.SK, sk_import_power, sk_import_reason, sk_export_power, sk_export_reason)

        capacity = TransferCapacity(utc_dt, [bc_atc_limit, sk_atc_limit])

        yield capacity


class TransferCapacity(object):
    '''Available Transfer Capacity Limit.'''

    def __init__(self, t, links):
        self._t = t
        self._links = links

    @property
    def t(self):
        '''Time (:class:`datetime.datetime`) at which intertie link limits are valid.
        '''
        return self._t


    @property
    def links(self):
        '''List of :class:`AtcLimit` objects.  List length is equal to
        the number of values in the :class:`AtcLink` enumeration.
        Presently :class:`AtcLink` has two values: Saskatchewan and
        BC (pyaeso 0.6, 2010-11-14).'''

        return self._links


class AtcLimit(object):
    '''Represents transfer limit for an intertie.'''

    def __init__(self, id, import_power, import_reason, export_power, export_reason):
        self._id = id
        self._import_power = Decimal(import_power)
        self._import_reason = import_reason
        self._export_power = Decimal(export_power)
        self._export_reason = export_reason

    @property
    def id(self):
        '''Value from :class:`AtcLink` enumeration.'''
        return self._id

    @property
    def import_power(self):
        '''Imported power (:class:`decimal.Decimal`).'''
        return self._import_power

    @property
    def import_reason(self):
        ''':class:`str` property.'''
        return self._import_reason

    @property
    def export_power(self):
        ''':class:`decimal.Decimal` property.'''
        return self._export_power

    @property
    def export_reason(self):
        ''':class:`str` property.'''
        return self._export_reason




def _normalize_atc_dtstr_to_utc(cells):
    # Sample line
    # 2009-01-01,1,330,"OPP 521 Limit",525,"Opp 312 Limit",35,"",153,""
    DT_FORMAT = "%Y-%m-%d"

    # Construct a naive datetime object
    is_dst = None

    # This series is to support Python < 2.5
    struct_time = strptime(cells[0], DT_FORMAT)
    timestamp = mktime(struct_time)
    t = datetime.fromtimestamp(timestamp)

    # Fiddle with "hours"
    hour = int(cells[1])
    if 1 <= hour and hour <= 24:
        t = t + timedelta(hours = hour)
    elif hour == 25:
        t = t + timedelta(hours = 2)
        is_dst = False
    else:
        #raise ValueError('hour {0} is invalid.'.format(cells))
        raise ValueError('hour ' + repr(cells) + ' is invalid.')

    # On a DST -> Standard time shift ATC report counts hours so:
    # 24 1 2 25 3 4 5 6
    #
    # Need to convert this to:
    #
    # 0* 1* 1 2 3 4 5 6
    #
    # Where * denotes is_dst flag set to True.

    # On a Standard time -> DST time shift ATC report counts hours so:
    # 24 1 2 4 5 6
    #
    # Need to convert this to:
    #
    # 0 1 3* 4* 5* 6*
    #
    # Where * denotes is_dst flag set to True.

    # Localize naive datetime objects.
    t_minus_one = t - timedelta(hours = 1)
    try:
        ab_dt = None
        AB_TZ.localize(t_minus_one, is_dst = is_dst)
    except pytz.AmbiguousTimeError:
        # one-hour after dst transition
        ab_dt = AB_TZ.localize(t_minus_one, is_dst = False)
    else:
        try:
            ab_dt = AB_TZ.localize(t, is_dst = is_dst)
        except pytz.AmbiguousTimeError:
            try:
                ab_dt = AB_TZ.localize(t, is_dst = True)
            except IndexError:
                print t
                raise
        except pytz.NonExistentTimeError:
            ab_dt = AB_TZ.localize(t + timedelta(hours = 1), is_dst = True)

    # convert from Alberta time to UTC time
    return pytz.utc.normalize(ab_dt.astimezone(pytz.utc))


class AtcLink(object):
    '''Enumeration of import/export transmission links.
    '''

    #: Transfer link to British Columbia.
    BC = 'BC'

    #: Transfer link to Saskatchewan.
    SK = 'SK'






#~ pyaeso is a python package that makes access to the Alberta, Canada's
#~ Electric System Operator's (AESO) Energy Trading System (ETS) easier.

#~ Copyright (C) 2010 Keegan Callin

#~ This program is free software: you can redistribute it and/or modify
#~ it under the terms of the GNU General Public License as published by
#~ the Free Software Foundation, either version 3 of the License, or
#~ (at your option) any later version.

#~ This program is distributed in the hope that it will be useful,
#~ but WITHOUT ANY WARRANTY; without even the implied warranty of
#~ MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#~ GNU General Public License for more details.

#~ You should have received a copy of the GNU General Public License
#~ along with this program.  If not, see
#~ <http://www.gnu.org/licenses/gpl-3.0.html>.

'''Access to the Alberta Interconnected Electric System Event log.  The
raw log is accessible at <http://ets.aeso.ca/ets_web/ip/Market/Reports/RealTimeShiftReportServlet>.

.. warning::
  This module is deprecated and slated for removal.  Please use
  :mod:`aeso.eventlog` instead.
'''


import warnings

warnings.warn('Module aeso.aieslog is deprecated; please use aeso.eventlog instead.', DeprecationWarning, stacklevel=2)

from eventlog import urlopen
from eventlog import parse_eventlog_file as parse_aieslog_file



# Standard library imports
import unittest
import sys

# Custom test libraries
import libtest

# Testable libraries
import test_examples
import test_ets
import test_marginal_pool_price
import test_csd
import test_aieslog
import test_eventlog
import test_util
import test_mpp
import test_atc
import test_asset

test_modules = (
    test_aieslog,
    test_eventlog,
    test_csd,
    test_util,
    test_mpp,
    test_atc,
    test_asset,
)

loader = unittest.TestLoader()
remote_suites = [
    loader.loadTestsFromModule(test_examples),
    loader.loadTestsFromTestCase(test_marginal_pool_price.TestMarginalPoolPriceServiceBehaviour),
    loader.loadTestsFromModule(test_ets.TestLiveEtsConnection),
]

for module in test_modules:
    try:
        remote_suites.extend(module.remote_suites)
    except AttributeError:
        pass

local_suites = [
    loader.loadTestsFromTestCase(test_marginal_pool_price.TestMarginalPoolPrice),
    loader.loadTestsFromTestCase(test_ets.TestDayBlockIt),
    loader.loadTestsFromTestCase(test_ets.TestPoolPrice),
    loader.loadTestsFromTestCase(test_ets.TestAssetList),
]

for module in test_modules:
    try:
        local_suites.extend(module.local_suites)
    except AttributeError:
        pass

if __name__ == '__main__':
    sys.exit(libtest.main(sys.argv[1:], local_suites, remote_suites))



'''Helper functions to access AESO's raw ETS pool price report as
published by AESO at
<http://ets.aeso.ca/ets_web/ip/Market/Reports/HistoricalPoolPriceReportServlet>.'''

########################################################################
## Standard library imports
import urllib
import urllib2
from datetime import date
from datetime import datetime
from datetime import timedelta
import shutil
import time
import csv
import decimal
from decimal import Decimal
from time import strptime
from time import mktime
import re


########################################################################
## Import 3rd party modules
import pytz


########################################################################
## AESO Modules
from aeso import AB_TZ
from aeso._util import DayBlockIt

_RE_DATEHOUR = re.compile('(\d+)/(\d+)/(\d+) (\d+)$')

def urlopen(start_date, end_date):
    '''Returns a file-like object attached to the ETS pool price report
    webservice.  Note that the webservice limits the number of days
    that can be queried to 721 days (as of 2009-11-12).

    :param start_date: :class:`datetime.date`
    :param end_date: :class:`datetime.date`
    :rtype: file-like object as returned by urlopen.

    Usage example::

        >>> # 3rd Party Libraries
        >>> from aeso import equilibrium
        >>> from datetime import date
        >>> from datetime import timedelta
        >>>
        >>> end_date = date.today()
        >>> start_date = end_date - timedelta(1)
        >>>
        >>> f = equilibrium.urlopen(start_date, end_date)
        >>> text = f.read()
        >>> f.close()
    '''
    DATE_FORMAT = '%m%d%Y'

    url = 'http://ets.aeso.ca/ets_web/ip/Market/Reports/HistoricalPoolPriceReportServlet'
    parameters = {
        'contentType' : 'csv',
        'beginDate' : start_date.strftime(DATE_FORMAT),
        'endDate' : end_date.strftime(DATE_FORMAT),
    }

    encoded_params = urllib.urlencode(parameters)
    #http://ets.aeso.ca/ets_web/ip/Market/Reports/HistoricalPoolPriceReportServlet?contentType=html&beginDate=08012009&endDate=08112009
    f = urllib2.urlopen(url, encoded_params)

    return f


def dump_equilibrium(f_out, start_date = date(1995, 1, 1), end_date = None):
    '''Downloads market equilibrium data from ETS and writes it to the
    file object f_out.  Unlike urlopen_pool_price there is no
    restriction on the amount of data that can be requested.  Internally
    an iterator is used to query data in 721 day blocks before it is
    written to f_out.

    :param f: file-like object
    :param start_date: :class:`datetime.date`
    :param end_date: :class:`datetime.date` (default date.today() + timedelta(1))


    Usage example::

        >>> # 3rd Party Libraries
        >>> from datetime import date
        >>> from datetime import timedelta
        >>> from aeso import equilibrium
        >>>
        >>> end_date = date.today()
        >>> start_date = end_date - timedelta(1)
        >>>
        >>> f = open('pool_price_report.csv', 'w')
        >>> equilibrium.dump_equilibrium(f, start_date, end_date)
        >>> f.close()
    '''

    if end_date is None:
        end_date = date.today() + timedelta(1)

    for (start, end) in DayBlockIt(start_date, end_date, 721):
        # print 'From', start, 'to', end
        f_in = urlopen(start, end)
        shutil.copyfileobj(f_in, f_out)

        if end < end_date:
            time.sleep(10)


def _normalize_pool_price_dtstr_to_utc(datetime_str):
    # Sample line:
    # ['Date (HE)', 'Price ($)', '30Ravg ($)', 'System Demand (MW)']
    # ['08/10/2009 15', '67.36', '39.67', '8623.0']
    datetime_str = datetime_str.strip()
    starred = False

    # Construct a naive datetime object
    try:
        if datetime_str.endswith('*'):
            starred = True
            datetime_str = datetime_str[0:-1]

        # This series is to support Python < 2.5
        struct_time = strptime(datetime_str, "%m/%d/%Y %H")
        timestamp = mktime(struct_time)
        t = datetime.fromtimestamp(timestamp)

        # Python >= 2.5
        # t = datetime.datetime.strptime(datetime_str, "%m/%d/%Y %H")

        # This code segment verifies that the Python 2.4 code section
        # above is equivalent to the Python >= 2.5 version.
        #~ if datetime.datetime.strptime(datetime_str, "%m/%d/%Y %H") != t:
            #~ raise ValueError('Problem!')

    except ValueError:
        # Often receive a line like this.
        # ['07/31/2009 24', '36.78', '41.83', '7556.0']
        #
        # AESO clock goes from 1-24 hours instead of expected
        # range of 0-23.
        #
        # Compensating for this strangeness
        miscreant_str = datetime_str
        match = _RE_DATEHOUR.match(miscreant_str)
        if match:
            year = int(match.group(3))
            month = int(match.group(1))
            day = int(match.group(2))
            hour = int(match.group(4))

            if hour == 24:
                t = datetime(year, month, day, 0) + timedelta(1)
            else:
                raise
        else:
            raise

    # Localize naive datetime objects.
    #
    # On a Daylight-Savings-Time (DST) switch, AESO hours are counted:
    # 1, 2, 2*
    #
    # Need to translate this to:
    # 1 (with is_dst option set to True), 1 (with is_dst option set to False), 2
    #
    if t.hour == 2 and not starred:
        # Testing to see if this second-hour of the day occurs
        # after a DST transition
        dst_test_time = datetime(t.year, t.month, t.day, t.hour - 1)
        try:
            AB_TZ.localize(dst_test_time, is_dst = None)
        except pytz.AmbiguousTimeError:
            # This "2" occurs after a DST transition
            ab_dt = AB_TZ.localize(dst_test_time, is_dst = False)
        else:
            # This "2" does not occur after a DST transition; no is_dst necessary
            ab_dt = AB_TZ.localize(t, is_dst = None)
    else:
        try:
            ab_dt = AB_TZ.localize(t, is_dst = None)
        except pytz.AmbiguousTimeError:
            if t.hour == 1 and not starred:
                # First hour occurring before a DST jump.
                ab_dt = AB_TZ.localize(t, is_dst = True)
            else:
                raise

    # convert from Alberta time to UTC time
    return pytz.utc.normalize(ab_dt.astimezone(pytz.utc))


class QpPoint(object):
    '''Represents the market equilibrium price and quantity exchanged at
    a given point in time.'''

    def __init__(self, t, price, demand):
        self._t = t
        self._price = Decimal(price)
        self._demand = Decimal(demand)

    @property
    def t(self):
        ''':class:`datetime.datetime` property.'''
        return self._t

    @property
    def price(self):
        ''':class:`decimal.Decimal` property.'''
        return self._price

    @property
    def demand(self):
        ''':class:`decimal.Decimal` property.'''
        return self._demand

    @classmethod
    def from_csvline(cls, line):
        '''Extracts a QpPoint object from a sequece object like:
        ['08/10/2009 15', '67.36', '39.67', '8623.0'].
        '''
        # ['Date (HE)', 'Price ($)', '30Ravg ($)', 'System Demand (MW)']

        # Normalize time string to UTC time.
        t = _normalize_pool_price_dtstr_to_utc(line[0])
        price = Decimal(line[1])
        demand = Decimal(line[3])

        point = QpPoint(t, price, demand)

        return point


def parse_equilibrium_file(f):
    '''Yields :class:`QpPoint` objects as extracted from the open CSV
    file-object *f*.

    Usage example::

        >>> # Standard library imports
        >>> from StringIO import StringIO
        >>> from datetime import date
        >>> from datetime import timedelta
        >>>
        >>> # 3rd Party Libraries
        >>> import aeso
        >>> from aeso import equilibrium
        >>>
        >>> end_date = date.today()
        >>> start_date = end_date - timedelta(1)
        >>>
        >>> f = StringIO()
        >>> equilibrium.dump_equilibrium(f, start_date, end_date)
        >>> f.seek(0)
        >>> data = list(equilibrium.parse_equilibrium_file(f))
        >>> f.close()
        >>>
        >>> # Yesterday's market clearing price/demand points.
        >>> for d in data:
        ...   ab_time = AB_TZ.normalize(d.t.astimezone(aeso.AB_TZ))
        ...   # print '{0} ${1} {2}MW'.format(ab_time, d.price, d.demand)
    '''

    reader = csv.reader(f)
    for idx, line in enumerate(reader):
        # ['Date (HE)', 'Price ($)', '30Ravg ($)', 'System Demand (MW)']
        # ['08/10/2009 15', '67.36', '39.67', '8623.0']
        try:
            try:
                yield QpPoint.from_csvline(line)
            except ValueError, e:
                #~ if line[0].strip().endswith('*'):
                    #~ # Star exists in output (meaning is unknown)
                    #~ # ignore this point.
                    #~ pass
                if line[0].strip() == '' or \
                    line[0].strip() == 'Pool Price' or \
                    line[0].strip() == 'Date (HE)':
                    # Date string is empty.  This is a header line or
                    # blank line.
                    pass
                else:
                    raise
            except IndexError, e:
                # Ignore the line; it does not have the right number of cells.
                # It may, for example, be blank.
                pass
            except decimal.InvalidOperation, e:
                if line[1].strip() == '-':
                    # No price data available. Ignore point.
                    pass
                else:
                    raise
        except (ValueError, IndexError, decimal.InvalidOperation), e:
            #raise ValueError('Unable to parse line {0}: {1}'.format(idx, repr(line)))
            raise ValueError('Unable to parse line ' + str(idx) + ': ' + repr(line))





from datetime import timedelta

class DayBlockIt(object):
    '''Steps over blocks of days between two time periods.  Each call to
    next() will return a 2-tuple containing a start date and end date as
    far apart as is permitted by the /days/ parameter.

    .. versionadded:: 0.6

    Example::

        >>> from aeso._util import DayBlockIt
        >>> from datetime import date
        >>>
        >>> start_date = date(1995, 1, 1)
        >>> end_date = date(1995, 1, 10)
        >>>
        >>> it = DayBlockIt(start_date, end_date, 4)
        >>> for first, last in it:
        ...   print first, last
        ...
        1995-01-01 1995-01-04
        1995-01-05 1995-01-08
        1995-01-09 1995-01-10
    '''

    def __init__(self, start_date, end_date, days):
        '''Create an object that iterates blocks of start/end dates of length
        /days/.

        @param days:  maximum number of days in each step.

        @type start_date: datetime.date
        @type end_date: datetime.date
        @type days: int
        '''
        if days == 0:
            raise ValueError('DayBlockIt() arg 3 must not be zero')
        self._delta = timedelta(days)
        self._delta_less_one = timedelta(days - days / abs(days))
        self._now = start_date
        self._start_date = start_date
        self._end_date = end_date

    def __iter__(self):
        return self

    def next(self):
        if (self._start_date <= self._now and self._now <= self._end_date) or \
            (self._start_date >= self._now and self._now >= self._end_date):
            start_date = self._now
            end_date = self._now + self._delta_less_one

            self._now = self._now + self._delta
            if self._start_date <= self._end_date:
                if self._now > self._end_date:
                    end_date = self._end_date
            else:
                if self._now < self._end_date:
                    end_date = self._end_date

            return (start_date, end_date)
        else:
            raise StopIteration()



########################################################################
## Standard library imports
import unittest
from os.path import dirname
from os.path import join
import doctest
#from datetime import date
from datetime import datetime
#from datetime import timedelta
import sys
import libtest
from cStringIO import StringIO
import bz2

########################################################################
## AESO Modules
from aeso import AB_TZ
from aeso import UTC_TZ
from aeso import equilibrium


class TestEquilibrium(unittest.TestCase):
    def test_parse_pool_price_file(self):
        test_series_file = join(dirname(__file__), 'res', 'ets_testseries.csv.bz2')

        f = bz2.BZ2File(test_series_file)
        points = list(equilibrium.parse_equilibrium_file(f))
        self.assertEquals(len(points), 6728)
        f.close()

    def test_datetime_normalization(self):
        # Test DST handling
        lut = {
            # DST active
            "10/26/1996 00" : datetime(1996, 10, 26, 6),
            "10/26/1996 01" : datetime(1996, 10, 26, 7),
            "10/26/1996 02" : datetime(1996, 10, 26, 8),
            "10/26/1996 03" : datetime(1996, 10, 26, 9),

            # DST ends this day
            "10/27/1996 00" : datetime(1996, 10, 27, 6),
            "10/27/1996 01" : datetime(1996, 10, 27, 7),
            "10/27/1996 02" : datetime(1996, 10, 27, 8),
            "10/27/1996 02*" : datetime(1996, 10, 27, 9),
            "10/27/1996 03" : datetime(1996, 10, 27, 10),

            # DST inactive
            "10/28/1996 00" : datetime(1996, 10, 28, 7),
            "10/28/1996 01" : datetime(1996, 10, 28, 8),
            "10/28/1996 02" : datetime(1996, 10, 28, 9),
            "10/28/1996 03" : datetime(1996, 10, 28, 10),
        }
        for datetime_str, expected_utc_datetime in lut.items():
            actual_dt = equilibrium._normalize_pool_price_dtstr_to_utc(datetime_str)
            expected_dt = UTC_TZ.localize(expected_utc_datetime)
            self.assertEquals(actual_dt, expected_dt)

        # Test handling of 24
        # '03/02/2005 24' -> 2009-03-03 0
        self.assertEquals(equilibrium._normalize_pool_price_dtstr_to_utc('03/02/2005 24'), AB_TZ.localize(datetime(2005, 3, 3, 0)).astimezone(UTC_TZ))


class TestEquilibriumRemote(unittest.TestCase):
    def test_pool_price_connection(self):
        '''ETS-Coupled Test.'''
        f = StringIO()
        equilibrium.dump_equilibrium(f)
        f.seek(0)
        points = list(equilibrium.parse_equilibrium_file(f))
        f.close()


loader = unittest.TestLoader()
remote_suites = [
    loader.loadTestsFromTestCase(TestEquilibriumRemote),
    doctest.DocTestSuite(equilibrium),
]

local_suites = [
    loader.loadTestsFromTestCase(TestEquilibrium),
]


if __name__ == '__main__':
    sys.exit(libtest.main(sys.argv[1:], local_suites, remote_suites))



# -*- coding: utf-8 -*-
#~ pyaeso is a python package that makes access to the Alberta, Canada's
#~ Electric System Operator's (AESO) Energy Trading System (ETS) easier.

#~ Copyright (C) 2010 Keegan Callin

#~ This program is free software: you can redistribute it and/or modify
#~ it under the terms of the GNU General Public License as published by
#~ the Free Software Foundation, either version 3 of the License, or
#~ (at your option) any later version.

#~ This program is distributed in the hope that it will be useful,
#~ but WITHOUT ANY WARRANTY; without even the implied warranty of
#~ MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#~ GNU General Public License for more details.

#~ You should have received a copy of the GNU General Public License
#~ along with this program.  If not, see
#~ <http://www.gnu.org/licenses/gpl-3.0.html>.

'''Access to the Alberta Interconnected Electric System Event log.  The
raw log is accessible at
<http://ets.aeso.ca/ets_web/ip/Market/Reports/RealTimeShiftReportServlet>.
'''

# Standard library imports
import urllib2
import csv
from time import strptime
from time import mktime
from datetime import datetime

# Custom libraries
from aeso import AB_TZ
import pytz

def urlopen():
    '''Returns an open file-object connected to AESO's Alberta
    Interconnected Electric System (AIES) event log webservice.

    .. versionadded:: 0.6
    '''

    src = 'http://ets.aeso.ca/ets_web/ip/Market/Reports/RealTimeShiftReportServlet?contentType=csv'
    return urllib2.urlopen(src)


class LogEntry(object):
    '''Class representing an entry in the Alberta Integrated Electric
    System Log.  This object will iterate over its members returning
    self.dt, and self.description in succession.

    .. versionadded:: 0.6

    Example Usage::
        >>> from aeso import eventlog
        >>> from datetime import datetime
        >>> dt = datetime(2010, 4, 22)
        >>> e = eventlog.LogEntry(dt, 'Sample log entry')
        >>> e.dt
        datetime.datetime(2010, 4, 22, 0, 0)
        >>> e.description
        'Sample log entry'
        >>> extracted_dt, extracted_desc = e
        >>> print extracted_dt
        2010-04-22 00:00:00
        >>> print extracted_desc
        Sample log entry
    '''

    def __init__(self, datetime, description):
        self.__dt = datetime
        self.__desc = description
        self.__tuple = (self.__dt, self.__desc)

    @property
    def dt(self):
        ''':class:`datetime.datetime` property.'''
        return self.__dt

    @property
    def description(self):
        ''':class:`str` property.'''
        return self.__desc

    def __iter__(self):
        return iter(self.__tuple)

    def __len__(self):
        return len(self.__tuple)

    def __getitem__(self, idx):
        return self.__tuple[idx]

    def __eq__(self, other):
        if self.__tuple == other.__tuple:
            return True
        else:
            return False


def parse_eventlog_file(f):
    '''Yields (:class:`LogEntry`, str) objects containing
    event datetime and description as extracted from file-like object
    `f`.  As always with pyaeso, datetimes are UTC offset-aware and
    should be converted to localized datetimes before being displayed
    to the user.

    .. versionadded:: 0.6

    Example Usage::
        >>> from aeso import eventlog
        >>> from aeso import AB_TZ
        >>>
        >>> from datetime import datetime
        >>> f = eventlog.urlopen()
        >>> for utc_dt, msg in eventlog.parse_eventlog_file(f):
        ...     # Convert UTC to Alberta timezone before printing
        ...     ab_dt = AB_TZ.normalize(utc_dt.astimezone(AB_TZ))
        ...     assert type(ab_dt) == datetime
        ...     assert type(msg) == str # Event description
        ...
        >>> f.close()
    '''

    num_extracted_rows = 0
    reader = csv.reader(f)
    for idx, cells in enumerate(reader):
        try:
            if len(cells) == 0:
                # Blank line.  Ignore and continue
                pass
            elif len(cells) == 2:
                dt_str = cells[0]
                entry = cells[1]

                struct_time = strptime(dt_str, "%m/%d/%Y %H:%M")
                timestamp = mktime(struct_time)
                dt = datetime.fromtimestamp(timestamp)
                ab_dt = AB_TZ.localize(dt)
                utc_dt = pytz.utc.normalize(ab_dt.astimezone(pytz.utc))

                num_extracted_rows += 1
                yield LogEntry(utc_dt, entry)
            else:
                raise IndexError('Incorrect number of cells.')
        except (IndexError, ValueError), e:
            if num_extracted_rows > 0:
                raise
            else:
                continue



########################################################################
## Standard library imports
from datetime import datetime
from datetime import date
from datetime import timedelta
import unittest
import doctest
import sys
import os
import bz2
from cStringIO import StringIO


########################################################################
## AESO Modules
from aeso import UTC_TZ
from aeso import atc
import libtest


class TestAtcLimits(unittest.TestCase):
    def test_date_parsing(self):
        converstions = [
            # Standard time -> daylight saavings time transition
            (('2000-04-01', '24'), datetime(2000, 4, 2, 7)),
            (('2000-04-02', '1'), datetime(2000, 4, 2, 8)),
            (('2000-04-02', '2'), datetime(2000, 4, 2, 9)),
            (('2000-04-02', '4'), datetime(2000, 4, 2, 10)),
            (('2000-04-02', '5'), datetime(2000, 4, 2, 11)),

            # Daylight-savings time -> standard time transition
            (('2009-10-31', '23'), datetime(2009, 11, 1, 5)),
            (('2009-10-31', '24'), datetime(2009, 11, 1, 6)),
            (('2009-11-01', '1'), datetime(2009, 11, 1, 7)),
            (('2009-11-01', '2'), datetime(2009, 11, 1, 8)),
            (('2009-11-01', '25'), datetime(2009, 11, 1, 9)),
            (('2009-11-01', '3'), datetime(2009, 11, 1, 10)),
            (('2009-11-01', '4'), datetime(2009, 11, 1, 11)),
        ]

        for cells, expected in converstions:
            expected = UTC_TZ.localize(expected)
            actual = atc._normalize_atc_dtstr_to_utc(cells)
            self.assertEquals(actual, expected)

    def test_parse_atc_file(self):
        test_series_file = os.path.join(os.path.dirname(__file__), 'res', 'atc_report.csv.bz2')

        atc_points = []
        f = bz2.BZ2File(test_series_file)
        atc_points.extend(atc.parse_atc_file(f))
        f.close()


class TestAtcLimitRemote(unittest.TestCase):
    def test_atc_connection(self):
        start_date = date(1999, 12, 22)
        end_date = date.today() + timedelta(1)

        f = StringIO()
        atc.dump_atc(f, start_date, end_date)
        f.seek(0)
        points = list(atc.parse_atc_file(f))
        f.close()


loader = unittest.TestLoader()
remote_suites = [
    loader.loadTestsFromTestCase(TestAtcLimitRemote),
    doctest.DocTestSuite(atc),
]

local_suites = [
    loader.loadTestsFromTestCase(TestAtcLimits),
]


if __name__ == '__main__':
    sys.exit(libtest.main(sys.argv[1:], local_suites, remote_suites))



'''Helpers to access AESO's asset list at
<http://ets.aeso.ca/ets_web/ip/Market/Reports/AssetListReportServlet>.
'''

########################################################################
## Standard library imports
import csv
import shutil
import re
import urllib
import urllib2


def urlopen_asset_list():
    '''Returns a file-like object containing data returned by the ETS
    asset list webservice.

    :rtype: file-like object as returned by urlopen.

    .. versionadded:: 0.6

    Usage example::

        >>> # 3rd Party Libraries
        >>> from aeso import asset
        >>>
        >>> f = asset.urlopen_asset_list()
        >>> text = f.read()
        >>> f.close()


    .. note::

        The raw ETS asset list report can be accessed at
        <http://ets.aeso.ca/ets_web/ip/Market/Reports/AssetListReportServlet>.
    '''

    url = 'http://ets.aeso.ca/ets_web/ip/Market/Reports/AssetListReportServlet'
    parameters = {
        'contentType' : 'csv',
    }

    encoded_params = urllib.urlencode(parameters)
    #http://ets.aeso.ca/ets_web/ip/Market/Reports/AssetListReportServlet?contentType=html
    f = urllib2.urlopen(url, encoded_params)

    return f


def dump_asset_list(f_out):
    '''Downloads asset list report and writes it to file-object *f*.

    .. versionadded:: 0.6

    Usage example::

        >>> # 3rd Party Libraries
        >>> from aeso import asset
        >>>
        >>> f = open('asset_list_report.csv', 'w')
        >>> asset.dump_asset_list(f)
        >>> f.close()
    '''

    f_in = urlopen_asset_list()
    shutil.copyfileobj(f_in, f_out)


_RE_DATEHOUR = re.compile('(\d+)/(\d+)/(\d+) (\d+)$')


class AssetType(object):
    '''Asset type enumeration.

    .. versionadded:: 0.6
    '''

    #:
    SOURCE = 'source'

    #:
    SINK = 'sink'

    _lut = {
        SOURCE : 'source',
        SINK : 'sink',
    }

    @classmethod
    def from_str(klass, string):
        '''Converts some simple strings to enumeration values.'''
        normalized = string.strip().lower()

        try:
            return klass._lut[normalized]
        except KeyError:
            #raise ValueError('Unknown asset type {0}.'.format(repr(string)))
            raise ValueError('Unknown asset type ' + repr(string))


class AssetStatus(object):
    '''Asset state enumeration.

    .. versionadded:: 0.6
    '''

    #:
    ACTIVE = 'active'

    #:
    INACTIVE = 'inactive'

    #:
    RETIRED = 'retired'

    #:
    SUSPENDED = 'suspended'

    _lut = {
        ACTIVE : 'active',
        INACTIVE : 'inactive',
        RETIRED : 'retired',
        SUSPENDED : 'suspended',
    }

    @classmethod
    def from_str(klass, string):
        '''Converts some simple strings to enumeration values.'''
        normalized = string.strip().lower()

        try:
            return klass._lut[normalized]
        except KeyError:
            raise ValueError('Unknown asset status ' + repr(string))
            #raise ValueError('Unknown asset status {0}.'.format(repr(string)))


_RE_ASSETNAME = re.compile('<[^>]*>\s*<[^>]*>(.*)')
def _normalize_asset_name(string):
    #'<A NAME=3Anchor"></A>301A 3070 Ret #1'
    match = _RE_ASSETNAME.match(string)
    if match:
        return match.group(1)
    else:
        return string


class Asset(object):
    '''Represents an asset be it an :class:`AssetType.SINK` and :class:`AssetType.SOURCE`

    .. versionadded:: 0.6
    '''

    def __init__(self, asset_name, asset_id, asset_type, status, participant_name, participant_id):
        self._asset_name = asset_name
        self._asset_id = asset_name
        self._asset_type = asset_type
        self._status = status
        self._participant_name = participant_name
        self._participant_id = participant_id

    @property
    def asset_name(self):
        ''':class:`str` property.'''
        return self._asset_name

    @property
    def asset_id(self):
        ''':class:`str` property.'''
        return self._asset_id

    @property
    def asset_type(self):
        ''':class:`AssetType` property.'''
        return self._asset_type

    @property
    def status(self):
        ''':class:`AssetStatus` property.'''
        return self._status

    @property
    def participant_name(self):
        ''':class:`str` property.'''
        return self._participant_name

    @property
    def participant_id(self):
        ''':class:`str` property.'''
        return self._participant_id


def parse_asset_list_file(f):
    '''Yields Asset objects extracted from the open file-object *f*.

    .. versionadded:: 0.6

    Usage example::

        >>> # 3rd Party Libraries
        >>> from aeso import asset
        >>>
        >>>
        >>> f = asset.urlopen_asset_list()
        >>> assets = list(asset.parse_asset_list_file(f))
        >>> f.close()'''
    reader = csv.reader(f)
    for idx, line in enumerate(reader):
        # ["Williams Lk Gen St - BCH","IPI1","Source","Retired","Inland Pacific Energy Services","IPES"]
        try:
            if idx > 2 and len(line) > 0:
                yield Asset(_normalize_asset_name(line[0]), line[1], AssetType.from_str(line[2]), AssetStatus.from_str(line[3]), line[4], line[5])
        except IndexError:
            # raised when number of cells in row is incorrect.
            #raise ValueError('Unable to parse line {0}: {1}'.format(idx, repr(line)))
            raise ValueError('Unable to parse line ' + str(idx) + ': ' + repr(line))




########################################################################
## Standard library imports
from datetime import date
import unittest
import doctest
import sys

########################################################################
## Custom Imports
import libtest
import aeso._util
from aeso._util import DayBlockIt

class TestDayBlockIt(unittest.TestCase):
    def test_iteration(self):
        start_date = date(1995, 1, 1)
        end_date = date(1995, 1, 10)

        it = DayBlockIt(start_date, end_date, 10)
        self.assertEquals(it.next(), (start_date, end_date))
        self.assertRaises(StopIteration, it.next)

        it = DayBlockIt(start_date, end_date, 5)
        self.assertEquals(it.next(), (date(1995, 1, 1), date(1995, 1, 5)))
        self.assertEquals(it.next(), (date(1995, 1, 6), date(1995, 1, 10)))
        self.assertRaises(StopIteration, it.next)


    def test_negative(self):
        start_date = date(1995, 1, 10)
        end_date = date(1995, 1, 1)

        it = DayBlockIt(start_date, end_date, -10)
        self.assertEquals(it.next(), (start_date, end_date))

        it = DayBlockIt(start_date, end_date, -5)
        self.assertEquals(it.next(), (date(1995, 1, 10), date(1995, 1, 6)))
        self.assertEquals(it.next(), (date(1995, 1, 5), date(1995, 1, 1)))
        self.assertRaises(StopIteration, it.next)


loader = unittest.TestLoader()
remote_suites = []

local_suites = [
    loader.loadTestsFromTestCase(TestDayBlockIt),
    doctest.DocTestSuite(aeso._util),
]


if __name__ == '__main__':
    sys.exit(libtest.main(sys.argv[1:], local_suites, remote_suites))



#~ pyaeso is a python package that makes access to the Alberta, Canada's
#~ Electric System Operator's (AESO) Energy Trading System (ETS) easier.

#~ Copyright (C) 2010 Keegan Callin

#~ This program is free software: you can redistribute it and/or modify
#~ it under the terms of the GNU General Public License as published by
#~ the Free Software Foundation, either version 3 of the License, or
#~ (at your option) any later version.

#~ This program is distributed in the hope that it will be useful,
#~ but WITHOUT ANY WARRANTY; without even the implied warranty of
#~ MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#~ GNU General Public License for more details.

#~ You should have received a copy of the GNU General Public License
#~ along with this program.  If not, see
#~ <http://www.gnu.org/licenses/gpl-3.0.html>.

'''Access to current supply demand (CSD) data.  The raw report can be
accessed at <http://ets.aeso.ca/ets_web/ip/Market/Reports/CSDReportServlet>.'''

# Standard library imports
import urllib2
import csv
from time import strptime
from datetime import datetime
from decimal import Decimal

# Custom libraries
from aeso import AB_TZ
from aeso._util import DayBlockIt
import pytz

def urlopen():
    '''Returns an open file-object connected to AESO's Current Supply
    Demand (CSD) webservice.

    .. versionadded:: 0.5'''
    src = 'http://ets.aeso.ca/ets_web/ip/Market/Reports/CSDReportServlet?contentType=csv'
    return urllib2.urlopen(src)


def _normalize_unit(unit):
    try:
        s = unit[unit.rindex('>') + 1:]
        while s.endswith('"'):
            s = s[0:-1]

        return s
    except ValueError:
        return unit


class CsdStat(object):
    '''Superclass of SystemStat and UnitStat'''
    pass


class SystemStat(CsdStat):
    '''Class representing a system statistic.

    .. versionadded:: 0.6

    Example Usage::
        >>> from aeso import csd
        >>> from datetime import datetime
        >>> dt = datetime(2010, 4, 22)
        >>> stat = csd.SystemStat(dt, 'Sample System Statistic', 10)
        >>> stat.dt
        datetime.datetime(2010, 4, 22, 0, 0)
        >>> stat.name
        'Sample System Statistic'
        >>> stat.value
        10
        >>> dt, name, value = stat
        >>> assert stat.dt == dt
        >>> assert stat.name == name
        >>> assert stat.value == value
    '''

    def __init__(self, datetime, name, value):
        self.__dt = datetime
        self.__name = name
        self.__value = value
        self.__tuple = (self.__dt, self.__name, self.__value)

    @property
    def dt(self):
        '''Entry :class:`datetime.datetime`.'''
        return self.__dt

    @property
    def name(self):
        '''Unit name, :class:`str`.'''
        return self.__name

    @property
    def value(self):
        '''Value, polymorphic.'''
        return self.__value

    def __iter__(self):
        return iter(self.__tuple)

    def __len__(self):
        return len(self.__tuple)

    def __getitem__(self, idx):
        '''Method is present only to support old code.'''
        return self.__tuple[idx]


class UnitStat(CsdStat):
    '''Class representing an entry in the Alberta Integrated Electric
    System Log.  This object will iterate over its members returning
    self.dt, and self.description in succession.

    .. versionadded:: 0.6

    Example Usage::
        >>> from aeso import csd
        >>> from datetime import datetime
        >>> dt = datetime(2010, 4, 22)
        >>> stat = csd.UnitStat(dt, 'Sample Unit Statistic', 10, 5, 0)
        >>> stat.dt
        datetime.datetime(2010, 4, 22, 0, 0)
        >>> stat.name
        'Sample Unit Statistic'
        >>> stat.mcr
        10
        >>> stat.tng
        5
        >>> stat.dcr
        0
        >>> dt, name, mcr, tng, dcr = stat
        >>> assert stat.dt == dt
        >>> assert stat.name == name
        >>> assert stat.mcr == mcr
        >>> assert stat.tng == tng
        >>> assert stat.dcr == dcr
        '''

    def __init__(self, datetime, name, mcr, tng, dcr):
        self.__dt = datetime
        self.__name = name
        self.__mcr = mcr
        self.__tng = tng
        self.__dcr = dcr
        self.__tuple = (self.__dt, self.__name, self.__mcr, self.__tng, self.__dcr)


    @property
    def dt(self):
        '''Entry :class:`datetime.datetime`.'''
        return self.__dt


    @property
    def name(self):
        '''Unit name.'''
        return self.__name


    @property
    def mcr(self):
        '''Maximum continuous rating.'''
        return self.__mcr

    @property
    def tng(self):
        '''Total net generation.'''
        return self.__tng

    @property
    def dcr(self):
        '''Dispatched contingency reserve.'''
        return self.__dcr

    def __iter__(self):
        return iter(self.__tuple)

    def __len__(self):
        return len(self.__tuple)

    def __getitem__(self, idx):
        '''Method is present only to support old code.'''
        return self.__tuple[idx]


def parse_csd_file(f, reference_dt = None):
    '''Yields subclasses of :class:`CsdStat` objects containing
    information from file-like object `f` (as returned by
    :func:`urlopen`).   Subclasses of :class:`CsdStat` presently
    yielded include:

    * :class:`SystemStat`.
    * :class:`UnitStat`.

    AESO presently publishes system statistics with the names (2010-03-10):

    * Alberta Total Net Generation
    * Interchange
    * Alberta Internal Load (AIL)
    * Alberta Load Responsibility
    * Contingency Reserve Required
    * Dispatched Contingency Reserve (DCR)
    * Dispatched Contingency Reserve - Gen
    * Dispatched Contingency Reserve - Other
    * BC Interchange flow
    * SK Interchange flow

    These system statistics will be represented by yielded
    :class:`SystemStat` objects.

    Unit data yielded in :class:`UnitStat` objects may actually be
    pseudo-units; For example, AESO provides data on a "Coal" unit
    that represents the generation output of all coal plants in the
    province.

    .. versionchanged:: 0.5.1
        Now yields subclasses of :class:`CsdData` instead of 3-tuples
        and 5-tuples.  The change should be transparent because
        :class:`CsdData` subclasses have tuple behaviours like __len__,
        __iter__, and __getitem__.

    .. versionadded:: 0.5

    Example Usage::

        >>> from aeso import csd
        >>> from aeso import AB_TZ
        >>> systemstats = []
        >>> unitstats = []
        >>> f = csd.urlopen()
        >>> for datapoint in csd.parse_csd_file(f):
        ...     if isinstance(datapoint, csd.SystemStat):
        ...         systemstats.append(datapoint)
        ...     elif isinstance(datapoint, csd.UnitStat):
        ...         ab_dt = AB_TZ.normalize(datapoint.dt.astimezone(AB_TZ))
        ...         unitstats.append(datapoint)
        ...     else:
        ...         raise TypeError('Unexpected type')
        ...
        >>> f.close()
    '''

    utc_dt = None
    num_extracted_rows = 0
    reader = csv.reader(f)
    for idx, cells in enumerate(reader):
        try:
            if len(cells) == 0:
                # Blank line.  Ignore and continue
                pass
            elif len(cells) == 1:
                value = cells[0]
                if value.strip() == 'Current Supply Demand Report':
                    # Ignore report title
                    pass
                elif value.startswith('Last Update :'):
                    struct_time = strptime(cells[0], "Last Update : %b %d, %Y %H:%M")
                    dt = datetime(*struct_time[0:6])
                    #print cells[0], '=>', dt
                    try:
                        ab_dt = AB_TZ.localize(dt, is_dst = None)
                    except pytz.AmbiguousTimeError:
                        # Have reference time
                        if reference_dt is None:
                            raise
                        else:
                            with_dst_dt = AB_TZ.localize(dt, is_dst = True)
                            without_dst_dt = AB_TZ.localize(dt, is_dst = False)

                            if abs(without_dst_dt - reference_dt) <= abs(with_dst_dt - reference_dt):
                                ab_dt = AB_TZ.localize(dt, is_dst = False)
                            else:
                                ab_dt = AB_TZ.localize(dt, is_dst = True)

                    utc_dt = pytz.utc.normalize(ab_dt.astimezone(pytz.utc))
                else:
                    raise ValueError('Unexpected line data')

                num_extracted_rows += 1
            elif len(cells) == 2:
                if dt is None:
                    raise ValueError('No datetime set')
                unit = _normalize_unit(cells[0])

                if unit == 'TOTAL':
                    continue

                if cells[1] == '-':
                    value = ''
                else:
                    value = Decimal(cells[1])

                num_extracted_rows += 1
                yield SystemStat(utc_dt, unit, value)
            elif len(cells) == 4:
                if dt is None:
                    raise ValueError('No datetime set')

                unit = _normalize_unit(cells[0])
                if unit == 'TOTAL':
                    continue
                mcr = Decimal(cells[1]) # Maximum continuous rating
                tng = Decimal(cells[2]) # total net generation
                dcr = Decimal(cells[3]) # Dispatched (and Accepted) Contingency Reserve

                num_extracted_rows += 1
                yield UnitStat(utc_dt, unit, mcr, tng, dcr)
            else:
                raise IndexError('Incorrect number of cells.')
        except (IndexError, ValueError), e:
            if num_extracted_rows > 0:
                raise
            else:
                continue



import unittest
import warnings

from os.path import dirname
from os.path import join

import bz2
import sys

import doctest

# Custom libraries
import libtest
from aeso import AB_TZ


class WarningMessage(object):
    """Copied from 2.6 sourcecode so that this works in Python >= 2.4"""

    _WARNING_DETAILS = ("message", "category", "filename", "lineno", "file",
                        "line")

    def __init__(self, message, category, filename, lineno, file=None,
                    line=None):
        local_values = locals()
        for attr in self._WARNING_DETAILS:
            setattr(self, attr, local_values[attr])

        self._category_name = category.__name__

    def __str__(self):
        return ("{message : %r, category : %r, filename : %r, lineno : %s, "
                    "line : %r}" % (self.message, self._category_name,
                                    self.filename, self.lineno, self.line))

_warning_log = []
_saved_showarning = warnings.showwarning
_saved_filters = warnings.filters
def _showwarning(*args, **kwargs):
    global _warning_log
    _warning_log.append(WarningMessage(*args, **kwargs))
warnings.showwarning = _showwarning

warnings.simplefilter("always")

# Trigger a warning.
from aeso import aieslog

# Verify some things
assert len(_warning_log) == 1
assert issubclass(_warning_log[-1].category, DeprecationWarning)
assert "deprecated" in str(_warning_log[-1].message)

warnings.showwarning = _saved_showarning
warnings.filters = _saved_filters



class TestAiesLog(unittest.TestCase):
    def test_parse_asset_list_file(self):
        test_file = join(dirname(__file__), 'res', 'RealTimeShiftReportServlet.csv.bz2')
        f = bz2.BZ2File(test_file)

        num_rows = 0
        for dt, entry in aieslog.parse_aieslog_file(f):
                num_rows += 1
                #print dt.astimezone(AB_TZ), entry
        f.close()
        self.assertEquals(num_rows, 29)


class TestAiesLogWebservice(unittest.TestCase):
    def test_parse_asset_list_file(self):
        f = aieslog.urlopen()
        rows = list(aieslog.parse_aieslog_file(f))
        f.close()
        self.assertTrue(len(rows) > 5)


loader = unittest.TestLoader()
remote_suites = [
    loader.loadTestsFromTestCase(TestAiesLogWebservice),
    doctest.DocTestSuite(aieslog),
]

local_suites = [
    loader.loadTestsFromTestCase(TestAiesLog),
]


if __name__ == '__main__':
    sys.exit(libtest.main(sys.argv[1:], local_suites, remote_suites))



'''Access to marginal pool price (MPP) data.  The raw report can be
accessed at <http://ets.aeso.ca/ets_web/ip/Market/Reports/HistoricalSystemMarginalPriceReportServlet>.'''

########################################################################
## Import Standard Library Modules

import urllib
import urllib2
from time import strptime
from time import mktime
import decimal
from decimal import Decimal
from datetime import datetime
from datetime import timedelta
from datetime import date
import shutil
import csv

########################################################################
## Import 3rd Party Modules
import pytz

########################################################################
## Import AESO package modules
from aeso import AB_TZ
from _util import DayBlockIt

def urlopen(start_date, end_date):
    '''Returns a file-like object attached to the ETS marginal pool
    price report.  The report is limited by AESO to returning 31 days of
    information (2010-02-09).  The report will include data for
    start_date but not for end_date.  The earliest date for which
    marginal price information is available is 1999-10-01 (2010-02-10).
    *start_date* must be before *end_date*.

    :param start_date: :class:`datetime.date`
    :param end_date: :class:`datetime.date`
    :rtype: file-like object as returned by urlopen.

    .. versionadded:: 0.6

    Usage example::

        >>> # Standard library imports
        >>> from datetime import date
        >>> from datetime import timedelta
        >>>
        >>> # 3rd Party Libraries
        >>> from pyaeso.ets import urlopen_marginal_pool_price
        >>>
        >>> start_date = date(2010, 1, 1)
        >>> end_date = date(2010, 1, 31) + timedelta(1)
        >>> # If you want to include all days in january, must add 24 hours.
        >>> # Remember that there will be no information returned for
        >>> # *end_date* itself, only for dates prior to it!
        >>>
        >>> f = urlopen_marginal_pool_price(start_date, end_date)
        >>> text = f.read()
        >>> f.close()'''

    DATE_FORMAT = '%m%d%Y'
    url = 'http://ets.aeso.ca/ets_web/ip/Market/Reports/HistoricalSystemMarginalPriceReportServlet'
    parameters = {
        'contentType' : 'csv',
        'beginDate' : start_date.strftime(DATE_FORMAT),
        'endDate' : end_date.strftime(DATE_FORMAT),
    }

    encoded_params = urllib.urlencode(parameters)
    #http://ets.aeso.ca/ets_web/ip/Market/Reports/HistoricalSystemMarginalPriceReportServlet?beginDate=02012010&endDate=02092010&contentType=csv
    f = urllib2.urlopen(url, encoded_params)

    return f


def _marginal_pool_price_dt(cells):
    date, hour_hint = [s.strip() for s in cells[0].split()]
    if hour_hint.endswith("*"):
        hour_hint = hour_hint[:-1]
    hour_hint = int(hour_hint)
    time = cells[1].strip()
    price = Decimal(cells[2])

    is_dst = None
    if time.endswith('*'):
        is_dst = False
        time = time[:-1]

    dt_str = date + ' ' + time
    dt_str = dt_str.strip()

    add_one_day = False
    try:
        struct_time = strptime(dt_str, "%m/%d/%Y %H:%M")
    except ValueError:
        # ValueError: time data '01/09/2010 24:00' does not match
        # format '%m/%d/%Y %H:%M'
        #
        # This happens because there is no "24:00" in strptime.  It
        # should be "00:00" of the following day!
        dt_str = dt_str.replace('24:', '00:')
        if hour_hint == 24:
            add_one_day = True

        struct_time = strptime(dt_str, "%m/%d/%Y %H:%M")

    # Create naive datetime object
    timestamp = mktime(struct_time)
    naive_dt = datetime.fromtimestamp(timestamp)
    if add_one_day:
        naive_dt += timedelta(1)

    # Convert naive datetime object to UTC.
    try:
        ab_dt = AB_TZ.localize(naive_dt, is_dst = is_dst)
    except pytz.AmbiguousTimeError:
        ab_dt = AB_TZ.localize(naive_dt, is_dst = True)
    utc_dt = pytz.utc.normalize(ab_dt.astimezone(pytz.utc))

    return utc_dt


def _filter_mpp_headers(src, dst):
    for line in src:
        stripped_line = line.strip()

        if not stripped_line or stripped_line == 'Historical System Marginal Price' or stripped_line == 'Date (HE),Time,Price ($)':
            continue
        else:
            dst.write(line)


def dump_mpp(dst, start_date = date(1999, 10, 1), end_date = None):
    '''Downloads market marginal pool price data from ETS and writes it
    to file object *dst*.  Unlike urlopen_marginal_pool_price there is
    no limit on the amount of data that can be requested.  Internally
    an iterator is used to query data in 31 day blocks before it is
    written to *dst*.  Output is included for start_date but excludes
    data for end_date.

    :param dst: writeable file object
    :param start_date: :class:`datetime.date`
    :param end_date: :class:`datetime.date`

    .. versionadded:: 0.6

    Usage example::

        >>> # Standard library imports
        >>> from datetime import date
        >>> from datetime import timedelta
        >>> from StringIO import StringIO
        >>>
        >>> # 3rd Party Libraries
        >>> from aeso import mpp
        >>>
        >>> start_date = date(2010, 1, 1)
        >>> end_date = date(2010, 1, 31) + timedelta(1)
        >>> # Remember, no data will be returned on end_date itself!
        >>>
        >>> f = StringIO()
        >>> mpp.dump_mpp(f)
        >>> text = f.getvalue()
        >>> f.close()'''

    if end_date is None:
        end_date = date.today() + timedelta(1)

    if start_date > end_date:
        raise ValueError('start_date must be before end_date')

    first_file = True
    for block_end, block_start in DayBlockIt(end_date, start_date, -31):
        f = urlopen(block_start, block_end + timedelta(1))
        if first_file:
            shutil.copyfileobj(f, dst)
            first_file = False
        else:
            _filter_mpp_headers(f, dst)
        f.close()


class PPoint(object):
    '''A price at a given point in time.

    Since PPoint objects will iterate over their properties t, and
    price, they can be unpacked:

    >>> from datetime import datetime
    >>> point = PPoint(datetime(2010, 2, 12, 10, 36), '4.56')
    >>> t, price = point
    '''
    def __init__(self, dt, price):
        self.__dt = dt
        self._price = Decimal(price)
        self._iterable = None

    @property
    def t(self):
        ''':class:`datetime.datetime` property'''
        return self.dt

    @property
    def dt(self):
        '''Equal to property :class:`PPoint.t`'''
        return self.__dt

    @property
    def price(self):
        ''':class:`decimal.Decimal` property'''
        return self._price

    def __iter__(self):
        if self._iterable is None:
            self._iterable = (self.t, self.price)

        return iter(self._iterable)


def parse_mpp_file(f):
    '''Yields a :class:`PPoint` object for each price point in marginal
    pool price data report file-object *f*.  As always, times are UTC.

    .. versionadded:: 0.6

    Usage example::

        >>> # Standard library imports
        >>> from datetime import date
        >>> from datetime import timedelta
        >>> from StringIO import StringIO
        >>>
        >>> # 3rd Party Libraries
        >>> from aeso.mpp import dump_mpp, parse_mpp_file
        >>> from aeso import AB_TZ, UTC_TZ
        >>>
        >>> start_date = date(2010, 1, 1)
        >>> end_date = date(2010, 1, 31) + timedelta(1)
        >>> # Remember, no data will be returned on end_date itself!
        >>>
        >>> f = StringIO()
        >>> dump_mpp(f)
        >>> f.seek(0)
        >>>
        >>> points = list(parse_mpp_file(f))
        >>>
        >>> for pp in points:
        ...   # PPoint objects are iterable and can be unpacked!
        ...   utc_dt, price = pp
        ...   # print '%s, $%f' % (AB_TZ.normalize(utc_dt.astimezone(AB_TZ)), price)
        ...   # *time* is in UTC, so it must be converted to Alberta
        ...   # timezone before display.
    '''

    ##########################
    #~ SAMPLE FILE:
    ##########################
    #~ Historical System Marginal Price
    #~ Date (HE),Time,Price ($)
    #~
    #~ "01/09/2010 24","23:51","42.00"
    #~ "01/09/2010 24","23:16","44.98"
    #~ "01/09/2010 24","23:09","45.00"
    #~ "01/09/2010 24","23:07","48.22"
    #~ "01/09/2010 24","23:05","49.00"
    #~ "01/09/2010 24","23:00","52.98"
    #~ "01/09/2010 23","22:55","42.00"
    #~ "01/09/2010 23","22:43","43.86"
    #~ "01/09/2010 23","22:41","46.00"
    #~ "01/09/2010 23","22:37","51.98"


    first_valid_line = False
    dst = False
    reader = csv.reader(f)
    for idx, cells in enumerate(reader):
        try:
            NUM_EXPECTED_CELLS = 3
            if len(cells) == NUM_EXPECTED_CELLS:
                dt = _marginal_pool_price_dt(cells)
                price = Decimal(cells[2])

                first_valid_line = True
                yield PPoint(dt, price)
            elif first_valid_line and len(cells) != 0:
                #raise ValueError('found {1} cell(s) where {2} were expected ("{3}").'.format(idx, len(cells), NUM_EXPECTED_CELLS, cells))
                raise ValueError('found ' + str(len(cells)) + ' where ' + str(NUM_EXPECTED_CELLS) + ' were expected ("' + str(cells) + '").')
        except (decimal.InvalidOperation, ValueError), e:
            if first_valid_line:
                #raise ValueError('On row {0}, {1}'.format(idx, str(e)))
                raise ValueError('On row ' + str(idx) + ', ' + str(e))



########################################################################
## Standard library imports
#~ from datetime import datetime
#~ from datetime import date
#~ from datetime import timedelta
import unittest
import doctest
import sys
import os
import bz2
from cStringIO import StringIO


########################################################################
## Import 3rd party modules
#~ import pytz


########################################################################
## AESO Modules
from aeso import asset
import libtest


class TestAssetList(unittest.TestCase):
    def test_parse_asset_list_file(self):
        test_series_file = os.path.join(os.path.dirname(__file__), 'res', 'asset_list.csv.bz2')

        f = bz2.BZ2File(test_series_file)
        assets = list(asset.parse_asset_list_file(f))
        self.assertEquals(len(assets), 1862)
        for a in assets:
            for char in '<>':
                # ETS embeds HTML anchors in the asset name.  Test to
                # make sure they have been properly stripped out.
                self.assertTrue(char not in a.asset_name)
        f.close()


class TestAssetRemote(unittest.TestCase):
    def test_asset_list_connection(self):
        '''ETS-Coupled Test.'''
        f = StringIO()
        asset.dump_asset_list(f)
        assets = list(asset.parse_asset_list_file(f))
        f.close()


loader = unittest.TestLoader()
remote_suites = [
    loader.loadTestsFromTestCase(TestAssetRemote),
    doctest.DocTestSuite(asset),
]

local_suites = [
    loader.loadTestsFromTestCase(TestAssetList),
]


if __name__ == '__main__':
    sys.exit(libtest.main(sys.argv[1:], local_suites, remote_suites))
